/* FileRefs.groovy
 * May 2008 by John Green.
 *
 * Find and fix code developed on Windows so it has file references
 * which are portable to unix.
 * For RUN *.p, RUN *.w, and for include file references:
 * 1. Makes sure forward unix-style slashes '/' are used rather than
 *    Windows back-slashes '\'.
 * 2. Makes sure that letter-case in the RUN or include ref matches
 *    the letter-case of the actual file name.
 * 
 * 
 * --- USAGE ---
 *
 * o  NOTE that ProRefactor writes potentially large amounts of data,
 *    and that this script may use large amounts of CPU, memory, and disc space.
 *
 * o  This script requires Proparse 3.1C or later.
 *
 * o  You must have ProRefactor and Groovy scripting set up. See:
 *    http://www.oehive.org/prorefactor/standalone_jar_file
 *
 * o  Configure filerefs.properties in your current working directory.
 *
 * o  Run the script: groovy FileRefs.groovy
 *
 * -------------
 *
 * Copyright (c) 2008 Joanju Software (www.joanju.com). All rights reserved.
 * Released under the terms of the Eclipse Public License v1.0:
 * http://www.eclipse.org/legal/epl-v10.html
 */
import org.prorefactor.refactor.RefactorSession
import org.prorefactor.treeparser.ParseUnit
import org.prorefactor.core.JPNode
import org.prorefactor.core.TokenTypes as TT
import org.prorefactor.refactor.FileTarget
import org.prorefactor.core.TreeUtils
import org.prorefactor.refactor.FileStuff
import org.prorefactor.core.Pstring
import org.prorefactor.macrolevel.IncludeRef

class FileRefs implements Runnable {

	def currSourceFileNumChanges = 0
	def totalChangedFiles = 0
	def totalChanges = 0
	File currSourceFile
	FileBuffer fileBuffer
	// Work with FileTarget objects, to avoid duplicate conversion on
	// statements in include files referenced more than once.
	// Work from a map that is sorted by FileTarget, so that multiple
	// entries from any one file are all grouped together.
	def fileTargets = new TreeMap<FileTarget, String>()
	def inputDirCanonical = ''
	def outputDirCanonical = ''
	Logger logger
	ParseUnit parseUnit
	def prsession = RefactorSession.instance
	ScriptSupport support

	final String LINESEP = System.getProperty('line.separator')

	

	void run() {
		support = new ScriptSupport('filerefs.properties', 'filerefs.log')
		try {
			if (! support.initialize())
				return
			inputDirCanonical = support.inputDir.canonicalPath
			outputDirCanonical = support.outputDir.canonicalPath
			run2_fileList()
		} finally {
			support.close()
		}
	}


	void run2_fileList() {
		logger = support.logger

		for (File file in support.fileScanner) {
			def relativePath = support.relativePath(file)
			logger.log(relativePath + ' ')
			parseUnit = new ParseUnit(file)
			try {
				// Parse the source file.
				parseUnit.treeParser01()
			} catch (Exception e) {
				// Show and log parse errors.
				logger.logLine(e.message)
				continue
			}
			run3_oneCompileUnit()
			logger.newline()
		}
		if (currSourceFileNumChanges > 0)
			writeCurrent()

		logger.log("$totalChanges changes in $totalChangedFiles files.")
	}


	void run3_oneCompileUnit() {
		fileTargets.clear()
		def nodeArray = TreeUtils.nodeArray(parseUnit.topNode)
		for (JPNode node in nodeArray) {
			if (node.type == TT.RUN)
				examineRunNode(node)
		}
		examineIncludeReferences()
		fileTargets.each{ target, node ->
			convert(target, node)
		}
	}


	void convert(FileTarget fileTarget, String fixedName) {
		loadIfNotCurrent(fileTarget.filename)

		def useTarget = fileTarget
		def line = fileBuffer.getLine(fileTarget.line)

		if (line.length() < fileTarget.column) {
			logNoConvert(useTarget, fixedName)
			return
		}

		// We might have the line/column of a '{' character
		// for an include reference. We have to find the column
		// of the filename.
		if (line[fileTarget.column - 1] == '{') {
			useTarget = filenameForIncludeRef(fileBuffer, fileTarget)
			line = fileBuffer.getLine(useTarget.line)
		}

		// We are only changing character case, or changing '\' to '/',
		// so it's easy for us to test that we can replace old text with new.

		if (line.length() < useTarget.column - 1 + fixedName.length()) {
			logNoConvert(useTarget, fixedName)
			return
		}

		def start = useTarget.column - 1
		def end = start + fixedName.length()

		def oldText = line.substring(start, end)
		def oldNormalized = oldText.toLowerCase().replace('\\', '/')
		def newNormalized = fixedName.toLowerCase().replace('\\', '/')
		if (oldNormalized != newNormalized) {
			logNoConvert(useTarget, fixedName)
			return
		}

		def newLine = new StringBuffer(line)
		newLine.replace(start, end, fixedName)
		fileBuffer.setLine(useTarget.line, newLine.toString())

		currSourceFileNumChanges++
		totalChanges++
	}


	void examineIncludeReferences() {
		macroRefLoop:
		for (macroRef in parseUnit.macroSourceArray) {
			if (! (macroRef instanceof IncludeRef))
				continue
			// The top level source file is also an IncludeRef object... not applicable.
			if (macroRef.parent == null)
				continue
			def includeRef = (IncludeRef)macroRef
			def refText = includeRef.fileRefName
			def file = FileStuff.findFile(refText)
			def wantName = file.canonicalPath.replace('\\', '/')
			wantName = wantName.substring(wantName.length() - refText.length())
			if (refText != wantName) {
				// The includeRef.fileIndex got us the name used in {name}.
				// We need the parent includeRef to find out what was the source file
				// where the parser found the "{name}" line of text.
				def parent = includeRef.parent
				if (! (parent instanceof IncludeRef)) {
					reportPreprocessorMess(includeRef)
					continue macroRefLoop
				}
				def srcFileName = parseUnit.fileIndex[parent.fileIndex]
				// Only convert files that are in the input or output directory.
				def canon = new File(srcFileName).canonicalPath
				if ( !	(  canon.startsWith(inputDirCanonical)
						|| canon.startsWith(outputDirCanonical)
						)
					)
					continue macroRefLoop
				def fileTarget = new FileTarget(
						srcFileName,
						includeRef.refLine,
						includeRef.refColumn)
				if (! fileTargets.containsKey(fileTarget)) {
					fileTargets.put(fileTarget, wantName)
				}
			}
		}
	}


	void examineRunNode(JPNode runNode) {
		// Check that it's a RUN statement, and not some other RUN node.
		if (! runNode.isStateHead())
			return
		// We are only interested in vanilla RUN statements,
		// not RUN STORED-PROCEDURE or RUN SUPER.
		if (runNode.state2)
			return
		def nameNode = runNode.firstChild()
		if (nameNode.type == TT.VALUE) {
			// Process RUN VALUE("hard/coded/filename"). Why people do this is beyond me.
			nameNode = nameNode.firstChild().nextSibling()
			if (nameNode.type != TT.QSTRING)
				return
		}
		// Only source that are in the input or output directory.
		def srcFileName = nameNode.filename
		def canon = new File(srcFileName).canonicalPath
		if ( !	(  canon.startsWith(inputDirCanonical)
				|| canon.startsWith(outputDirCanonical)
				)
			)
			return
		def nameText = nameNode.text
		Pstring pstring = null
		if (nameText.startsWith('"') || nameText.startsWith("'")) {
			pstring = new Pstring(nameText)
			nameText = pstring.justText()
		}
		if (! (nameText.endsWith('.p') || nameText.endsWith('.w')))
			return
		def file = FileStuff.findFile(nameText)
		if (!file)
			return
		def wantName = file.canonicalPath.replace('\\', '/')
		if (! wantName.endsWith(nameText)) {
			def fileTarget = new FileTarget(srcFileName, nameNode.line, nameNode.column)
			if (! fileTargets.containsKey(fileTarget)) {
				wantName = wantName.substring(wantName.length() - nameText.length())
				if (pstring) {
					wantName = pstring.quote + wantName + pstring.quote
					if (pstring.attributes)
						wantName = wantName + pstring.attributes
				}
				fileTargets.put(fileTarget, wantName)
			}
		}
	}


	/** Get a target for an include ref's filename.
	 * Given an input FileTarget for an include ref '{' char,
	 * return a new FileTarget for the first character of the include file name.
	 * It might be on a different line than the '{' char.
	 * It might be the first character after a single or double quote char.
	 */
	FileTarget filenameForIncludeRef(FileBuffer fileBuffer, FileTarget fileTarget) {
		// Notes: We are working with line and column positions that count from one,
		// but when working with the line as an array of characters, we have to remember
		// that the first character is at array position zero.
		def newTarget = new FileTarget(
				fileTarget.filename,
				fileTarget.line,
				fileTarget.column + 1)
		line_loop:
		while (true) {
			def line = fileBuffer.getLine(newTarget.line)
			char_loop:
			while (true) {
				if (newTarget.column > line.length()) {
					newTarget.line++
					newTarget.column = 1
					continue line_loop
				}
				def c = line[newTarget.column - 1]
				if (c.trim() == '') {
					// skip over space and tab characters
					newTarget.column++
					continue char_loop
				}
				if (c=='"' || c=="'")
					newTarget.column++
				return newTarget
			}
		}
	}


	void loadIfNotCurrent(String filename) {
		// Find the (possibly relative) file name on the propath.
		def newfile = FileStuff.findFile(filename)
		if (newfile.equals(currSourceFile))
			return
		if (currSourceFileNumChanges > 0)
			writeCurrent()
		currSourceFileNumChanges = 0
		currSourceFile = newfile
		fileBuffer = new FileBuffer(newfile)
	}


	/** Report an instance that could not be converted. */
	void logNoConvert(FileTarget fileTarget, String fixedName) {
		logger.logLine(
			"Could not convert ${fileTarget.filename} line ${fileTarget.line}. Should reference: $fixedName"	
			)
	}


	/** Report a reference to an include file that comes from preprocessing.
	 * If the text of an include ref comes from a &GLOBAL or &LOCAL defined macro,
	 * then we don't attempt to fix it. We just report it. Ex:
	 * &GLOBAL-DEFINE MESS {myinclude.i}
	 * {&MESS} // this line is the problem.
	 */
	void reportPreprocessorMess(IncludeRef includeRef) {
		def intermediate = includeRef.parent
		def parent = intermediate.parent
		while (! (parent instanceof IncludeRef)) {
			intermediate = parent
			parent = intermediate.parent
		}
		def refdName = parseUnit.fileIndex[includeRef.fileIndex]
		def src = parseUnit.fileIndex[parent.fileIndex]
		def line = intermediate.refLine
		logger.logLine(
			"Could not refactor preprocessed reference to $refdName from $src line $line")
	}


	/** Write the current (modified) source file to the output directory. */
	void writeCurrent() {
		def srcfileCanonPath = currSourceFile.canonicalPath
		def outdirCanonPath = support.outputDir.canonicalPath
		File outFile
		if (srcfileCanonPath.startsWith(outdirCanonPath)) {
			// We are re-writing a file that's already in the output dir.
			outFile = currSourceFile
		} else {
			totalChangedFiles++
			outFile = new File(
				outdirCanonPath
				+ File.separator
				+ srcfileCanonPath.substring(support.convertCanonicalDir.length())
				)
		}
		fileBuffer.write(outFile)
	}


}
